
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  FAT.C
*   DESCRIP  :  FENRIS FAT Module
*   DATE     :  November 16, 1998
*
*
***************************************************************************/

#include "globals.h"

void NWLockFat(VOLUME *volume)
{
#if (LINUX_SLEEP)
    if (WaitOnSemaphore(&volume->FatSemaphore) == -EINTR)
       NWFSPrint("lock fat was interrupted\n");
#endif
}

void NWUnlockFat(VOLUME *volume)
{
#if (LINUX_SLEEP)
    SignalSemaphore(&volume->FatSemaphore);
#endif
}

MIRROR_LRU *RemoveFAT(VOLUME *volume, MIRROR_LRU *lru);
void InsertFAT(VOLUME *volume, MIRROR_LRU *lru);
void InsertFATTop(VOLUME *volume, MIRROR_LRU *lru);
ULONG FreeFAT(VOLUME *volume);

ULONG InitializeFAT_LRU(VOLUME *volume)
{
    volume->FATListHead = volume->FATListTail = 0;
    volume->FATListLock = volume->FATListBlocks = 0;

    volume->FATBlockHash = NWFSCacheAlloc(BLOCK_NUMBER_HASH_SIZE, FATHASH_TAG);
    if (!volume->FATBlockHash)
       return -1;

    volume->FATBlockHashLimit = NUMBER_OF_BLOCK_HASH_ENTRIES;
    NWFSSet(volume->FATBlockHash, 0, BLOCK_NUMBER_HASH_SIZE);

    return 0;
}

ULONG FreeFAT_LRU(VOLUME *volume)
{
    register MIRROR_LRU *lru;

    while (volume->FATListHead)
    {
       lru = RemoveFAT(volume, (MIRROR_LRU *)volume->FATListHead);
       if (lru)
       {
	  if (lru->BlockState == L_DIRTY)
	  {
	     if (lru->Cluster1 != (ULONG) -1)
	     {
		if (WriteJournalVolumeCluster(volume,
					    lru->Cluster1,
					    lru->CacheBuffer,
					    lru->ClusterSize,
					    FAT_PRIORITY))
		{
		   NWFSPrint("PANIC -- dirty fat LRU block #%u [%X] write failed\n",
			(unsigned int)lru->Cluster1,
			(unsigned int)lru->Cluster1);
		}
	     }

	     if (lru->Cluster2 != (ULONG) -1)
	     {
		if (WriteJournalVolumeCluster(volume,
					    lru->Cluster2,
					    lru->CacheBuffer,
					    lru->ClusterSize,
					    FAT_PRIORITY))
		{
		   NWFSPrint("PANIC -- dirty fat LRU mirror #%u [%X] write failed\n",
			(unsigned int)lru->Cluster2, (unsigned int)lru->Cluster2);
		}
	     }
	     lru->BlockState = L_DATAVALID;
	  }
	  NWFSFree(lru->CacheBuffer);
	  NWFSFree(lru);
       }
    }

    if (volume->FATBlockHash)
       NWFSFree(volume->FATBlockHash);
    volume->FATBlockHashLimit =0;
    volume->FATBlockHash = 0;

    return 0;

}

MIRROR_LRU *AllocateFAT_LRUElement(VOLUME *volume)
{
    register MIRROR_LRU *lru;

    lru = NWFSAlloc(sizeof(MIRROR_LRU), FATLRU_TAG);
    if (!lru)
       return 0;

    NWFSSet(lru, 0, sizeof(MIRROR_LRU));
    lru->BlockState = L_FREE;
    lru->Cluster1 = -1;
    lru->Cluster2 = -1;
    lru->ClusterSize = volume->ClusterSize;

    lru->CacheBuffer = NWFSCacheAlloc(volume->ClusterSize, FAT_TAG);
    if (!lru->CacheBuffer)
    {
       NWFSFree(lru);
       return 0;
    }
    NWFSSet(lru->CacheBuffer, 0, volume->ClusterSize);
    return lru;
}

void FreeFAT_LRUElement(MIRROR_LRU *lru)
{
    if (lru->CacheBuffer)
       NWFSFree(lru->CacheBuffer);
    NWFSFree(lru);
    return;
}

ULONG FlushFAT(VOLUME *volume)
{
    register MIRROR_LRU *lru;

    NWLockFat(volume);

    lru = volume->FATListHead;
    while (lru)
    {
       if (lru->BlockState == L_DIRTY)
       {

	  if (lru->Cluster1 != (ULONG) -1)
	  {
	     if (WriteJournalVolumeCluster(volume,
					 lru->Cluster1,
					 lru->CacheBuffer,
					 lru->ClusterSize,
					 FAT_PRIORITY))
		continue;
	  }

	  if (lru->Cluster2 != (ULONG) -1)
	  {
	     if (WriteJournalVolumeCluster(volume,
					 lru->Cluster2,
					 lru->CacheBuffer,
					 lru->ClusterSize,
					 FAT_PRIORITY))
		continue;
	  }
	  lru->BlockState = L_DATAVALID;
       }
       lru = lru->next;
    }

    NWUnlockFat(volume);
    return 0;
}

ULONG FlushFATBuffer(VOLUME *volume, MIRROR_LRU *lru)
{
    if (lru->BlockState == L_DIRTY)
    {
       if (lru->Cluster1 != (ULONG) -1)
       {
	  if (WriteJournalVolumeCluster(volume, lru->Cluster1,
				      lru->CacheBuffer, lru->ClusterSize,
				      FAT_PRIORITY))
	     return -1;
       }

       if (lru->Cluster2 != (ULONG) -1)
       {
	  if (WriteJournalVolumeCluster(volume, lru->Cluster2,
				      lru->CacheBuffer, lru->ClusterSize,
				      FAT_PRIORITY))
	     return -1;
       }
       lru->BlockState = L_DATAVALID;
    }
    return 0;
}


MIRROR_LRU *RemoveFAT(VOLUME *volume, MIRROR_LRU *lru)
{
    NWLockFat(volume);
    if (volume->FATListHead == lru)
    {
       volume->FATListHead = (void *) lru->next;
       if (volume->FATListHead)
	  volume->FATListHead->prior = NULL;
       else
	  volume->FATListTail = NULL;
    }
    else
    {
       lru->prior->next = lru->next;
       if (lru != volume->FATListTail)
	  lru->next->prior = lru->prior;
       else
	  volume->FATListTail = lru->prior;
    }
    if (volume->FATListBlocks)
       volume->FATListBlocks--;
    NWUnlockFat(volume);
    return lru;

}

void InsertFAT(VOLUME *volume, MIRROR_LRU *lru)
{
    NWLockFat(volume);
    if (!volume->FATListHead)
    {
       volume->FATListHead = lru;
       volume->FATListTail = lru;
       lru->next = lru->prior = 0;
    }
    else
    {
       volume->FATListTail->next = lru;
       lru->next = 0;
       lru->prior = volume->FATListTail;
       volume->FATListTail = lru;
    }
    volume->FATListBlocks++;
    NWUnlockFat(volume);
    return;

}

void InsertFATTop(VOLUME *volume, MIRROR_LRU *lru)
{
    NWLockFat(volume);
    if (!volume->FATListHead)
    {
       volume->FATListHead = lru;
       volume->FATListTail = lru;
       lru->next = lru->prior = 0;
    }
    else
    {
       lru->next = volume->FATListHead;
       lru->prior = 0;
       volume->FATListHead->prior = lru;
       volume->FATListHead = lru;
    }
    volume->FATListBlocks++;
    NWUnlockFat(volume);
    return;

}

MIRROR_LRU *CreateFatLRU(VOLUME *volume, ULONG cluster,
			 ULONG mirror, ULONG index)
{
    register MIRROR_LRU *lru;
    register ULONG retCode;

    lru = AllocateFAT_LRUElement(volume);
    if (lru)
    {
       retCode = ReadJournalVolumeCluster(volume, cluster, lru->CacheBuffer,
					   volume->ClusterSize,
					   FAT_PRIORITY);
       if (retCode)
       {
	  FreeFAT_LRUElement(lru);
	  return 0;
       }
       lru->ModLock = 0;
       lru->BlockState = L_DATAVALID;
       lru->Cluster1 = cluster;
       lru->Cluster2 = mirror;
       lru->BlockIndex = index;
       InsertFAT(volume, lru);

       retCode = AddToFATHash(volume, lru);
       if (retCode)
       {
	  RemoveFAT(volume, lru);
	  FreeFAT_LRUElement(lru);
	  return 0;
       }

       retCode = SetAssignedClusterValue(volume, cluster, 1);
       if (retCode)
       {
	  NWFSPrint("nwfs:  assigned bit block value not set cluster-%X\n",
		    (unsigned int)cluster);
       }

       retCode = SetAssignedClusterValue(volume, mirror, 1);
       if (retCode)
       {
	  NWFSPrint("nwfs:  assigned bit block value not set mirror-%X\n",
		    (unsigned int)mirror);
       }
       return lru;
    }
    return 0;

}

FAT_ENTRY *GetFatEntry(VOLUME *volume, ULONG Cluster)
{
    register LRU_HASH_LIST *HashTable;
    register FAT_ENTRY *Table;
    register MIRROR_LRU *lru;
    register ULONG hash;
    register ULONG EntriesPerCluster;
    register ULONG Block, Offset;

    if ((Cluster == (ULONG) -1) || (Cluster & 0x80000000))
       return 0;

    if (Cluster > (volume->MountedVolumeClusters - 1))
       return 0;

    EntriesPerCluster = volume->ClusterSize / sizeof(FAT_ENTRY);
    Block = Cluster / EntriesPerCluster;
    Offset = Cluster % EntriesPerCluster;

    hash = (Block & (volume->FATBlockHashLimit - 1));
    HashTable = (LRU_HASH_LIST *) volume->FATBlockHash;
    if (!HashTable)
    {
       return 0;
    }

    lru = (MIRROR_LRU *) HashTable[hash].head;
    while (lru)
    {
       if (lru->BlockIndex == Block)
       {
	  Table = (FAT_ENTRY *) lru->CacheBuffer;
	  return (FAT_ENTRY *) &Table[Offset];
       }
       lru = lru->hashNext;
    }
    return 0;

}

FAT_ENTRY *GetFatEntryAndLRU(VOLUME *volume, ULONG Cluster,
				 MIRROR_LRU **rlru)
{
    register LRU_HASH_LIST *HashTable;
    register FAT_ENTRY *Table;
    register MIRROR_LRU *lru;
    register ULONG hash;
    register ULONG EntriesPerCluster;
    register ULONG Block, Offset;

    if ((Cluster == (ULONG) -1) || (Cluster & 0x80000000))
       return 0;

    if (Cluster > (volume->MountedVolumeClusters - 1))
       return 0;

    if (rlru)
       *rlru = 0;

    EntriesPerCluster = volume->ClusterSize / sizeof(FAT_ENTRY);
    Block = Cluster / EntriesPerCluster;
    Offset = Cluster % EntriesPerCluster;

    hash = (Block & (volume->FATBlockHashLimit - 1));
    HashTable = (LRU_HASH_LIST *) volume->FATBlockHash;
    if (!HashTable)
    {
       return 0;
    }

    lru = (MIRROR_LRU *) HashTable[hash].head;
    while (lru)
    {
       if (lru->BlockIndex == Block)
       {
	  Table = (FAT_ENTRY *) lru->CacheBuffer;
	  if (rlru)
	     *rlru = lru;
	  return (FAT_ENTRY *) &Table[Offset];
       }
       lru = lru->hashNext;
    }
    return 0;

}

ULONG ReadFATTable(VOLUME *volume)
{
    register MIRROR_LRU *lru;
    register FAT_ENTRY *FAT1, *FAT2, *verify;
    register ULONG retCode;
    register long index1, index2;
    register ULONG Block = 0, cluster1, cluster2, i;
    register ULONG last_cluster1, last_cluster2;
    register BYTE *FATCopy;

    FATCopy = NWFSCacheAlloc(volume->ClusterSize, FAT_WORKSPACE_TAG);
    if (!FATCopy)
       return -3;

#if (MOUNT_VERBOSE)
    NWFSPrint("*** Reading FAT Tables ***\n");
#endif

    last_cluster1 = 0;
    last_cluster2 = 0;
    cluster1 = volume->FirstFAT;
    cluster2 = volume->SecondFAT;

    if (!cluster2)
    {
       NWFSPrint("ffat-%X  sfat-%x  fdir-%x  sdir-%X\n",
		 (unsigned int) volume->FirstFAT,
		 (unsigned int) volume->SecondFAT,
		 (unsigned int) volume->FirstDirectory,
		 (unsigned int) volume->SecondDirectory);

       NWFSPrint("nwfs:  second FAT is NULL\n");
       NWFSFree(FATCopy);
       return -1;
    }

    lru = CreateFatLRU(volume, cluster1, cluster2, Block);
    if (!lru)
    {
       NWFSPrint("nwfs:  could not allocate LRU element in ReadFATTables\n");
       NWFSFree(FATCopy);
       return -1;
    }

    retCode = ReadJournalVolumeCluster(volume, cluster2, FATCopy,
					volume->ClusterSize,
					FAT_PRIORITY);
    if (retCode)
    {
       NWFSPrint("nwfs:  could not read volume cluster in ReadFATTables\n");
       NWFSFree(FATCopy);
       return retCode;
    }

    FAT1 = GetFatEntry(volume, cluster1);
    FAT2 = GetFatEntry(volume, cluster2);
    if (!FAT1 || !FAT2)
    {
       NWFSPrint("nwfs:  error reading fat entry ReadFATTables\n");
       NWFSFree(FATCopy);
       return -1;
    }

    index1 = FAT1->FATIndex;
    index2 = FAT2->FATIndex;

    while (TRUE)
    {
       if (NWFSCompare(lru->CacheBuffer, FATCopy, volume->ClusterSize))
       {
	  extern void dumpRecordBytes(BYTE *, ULONG);
	  register BYTE *p = lru->CacheBuffer;

	  for (i=0; i < (volume->ClusterSize / 64); i++)
	  {
	     if (NWFSCompare(&p[i * 64], &FATCopy[i * 64], 64))
	     {
		NWFSPrint("nwfs:  FAT data mirror mismatch\n");
		NWFSPrint("offset into cluster is 0x%X (%u bytes)\n",
			  (unsigned)(i * 64), (unsigned)(i * 64));

		NWFSPrint("index1-%d cluster1-[%08X] next1-[%08X]\n", (int)index1,
		    (unsigned int)cluster1,
		    (unsigned int)FAT1->FATCluster);
		dumpRecordBytes(&p[i * 64], 64);

		NWFSPrint("index2-%d cluster2-[%08X] next2-[%08X]\n", (int)index2,
		    (unsigned int)cluster2,
		    (unsigned int)FAT2->FATCluster);
		dumpRecordBytes(&FATCopy[i * 64], 64);
		break;
	     }
	  }
	  NWFSFree(FATCopy);
	  return -3;
       }

       if (FAT1->FATCluster == (ULONG) -1)
	  break;

       if (FAT2->FATCluster == (ULONG) -1)
	  break;

       if (!FAT1->FATCluster)
       {
	  NWFSPrint("nwfs:  Free Cluster Detected in FAT1 Chain\n");
	  NWFSFree(FATCopy);
	  return -3;
       }

       if (!FAT2->FATCluster)
       {
	  NWFSPrint("nwfs:  Free Cluster Detected in FAT2 Chain\n");
	  NWFSFree(FATCopy);
	  return -3;
       }

       if (FAT1->FATCluster & 0x80000000)
       {
	  NWFSPrint("nwfs:  SubAlloc Node Detected in FAT1\n");
	  NWFSFree(FATCopy);
	  return -3;
       }

       if (FAT2->FATCluster & 0x80000000)
       {
	  NWFSPrint("nwfs:  SubAlloc Node Detected in FAT2\n");
	  NWFSFree(FATCopy);
	  return -3;
       }

       Block++;
       last_cluster1 = cluster1;
       last_cluster2 = cluster2;
       cluster1 = FAT1->FATCluster;
       cluster2 = FAT2->FATCluster;

       if (last_cluster1 > cluster1)
       {
	  NWFSPrint("nwfs:  FAT1 chain overlaps on itself\n");
	  NWFSFree(FATCopy);
	  return -1;
       }

       if (last_cluster2 > cluster2)
       {
	  NWFSPrint("nwfs:  FAT2 chain overlaps on itself\n");
	  NWFSFree(FATCopy);
	  return -1;
       }

       lru = CreateFatLRU(volume, cluster1, cluster2, Block);
       if (!lru)
       {
	  NWFSPrint("nwfs:  could not allocate LRU element in ReadFATTables\n");
	  NWFSFree(FATCopy);
	  return -1;
       }

       retCode = ReadJournalVolumeCluster(volume,
					   cluster2,
					   FATCopy,
					   volume->ClusterSize,
					   FAT_PRIORITY);
       if (retCode)
       {
	  NWFSPrint("nwfs:  Read Error FAT2 returned %d\n", (int)retCode);
	  NWFSFree(FATCopy);
	  return retCode;
       }

       FAT1 = GetFatEntry(volume, cluster1);
       FAT2 = GetFatEntry(volume, cluster2);
       if (!FAT1 || !FAT2)
       {
	  NWFSPrint("nwfs:  error reading fat entry ReadFATTables\n");
	  NWFSFree(FATCopy);
	  return -1;
       }

       if (index1 > FAT1->FATIndex)
       {
	  NWFSPrint("nwfs:  FAT1 index overlaps on itself\n");
	  NWFSFree(FATCopy);
	  return -1;
       }

       if (index2 > FAT2->FATIndex)
       {
	  NWFSPrint("nwfs:  FAT2 index overlaps on itself\n");
	  NWFSFree(FATCopy);
	  return -1;
       }

       index1 = FAT1->FATIndex;
       index2 = FAT2->FATIndex;

    }
    NWFSFree(FATCopy);

    if (FAT1->FATCluster != FAT2->FATCluster)
    {
       NWFSPrint("nwfs:  FAT Table chain mirror mismatch\n");
       return -3;
    }

#if (MOUNT_VERBOSE)
    NWFSPrint("*** Verifying FAT Tables ***\n");
#endif

    for (i=0; i < volume->MountedVolumeClusters; i++)
    {
       verify = GetFatEntry(volume, i);
       if (verify)
       {
	  if (verify->FATCluster)
	  {
	     retCode = SetFreeClusterValue(volume, i, 1);  // block is allocated
	     if (retCode)
	     {
		NWFSPrint("nwfs:  bit block value not set cluster-%X [1]\n",
		       (unsigned int)i);
	     }

	     // this is a sanity check for ECC memory bit errors
	     if (!GetFreeClusterValue(volume, i))
	     {
		NWFSPrint("nwfs:  bit block value incorrect cluster-%X [1]\n",
		       (unsigned int)i);
	     }
	     volume->VolumeAllocatedClusters++;
	  }
	  else
	  {
	     retCode = SetFreeClusterValue(volume, i, 0);  // block is free
	     if (retCode)
	     {
		NWFSPrint("nwfs:  bit block value not set cluster-%X [0]\n",
		       (unsigned int)i);
	     }

	     // this is a sanity check for ECC memory bit errors
	     if (GetFreeClusterValue(volume, i))
	     {
		NWFSPrint("nwfs:  bit block value incorrect cluster-%X [0]\n",
		       (unsigned int)i);
	     }
	     volume->VolumeFreeClusters++;
	  }
       }
       else
       {
	  NWFSPrint("nwfs:  FAT Table data errors detected\n");
	  return -1;
       }
    }

    return 0;

}

ULONG AddToFATHash(VOLUME *volume, MIRROR_LRU *lru)
{
    register ULONG hash;
    register LRU_HASH_LIST *HashTable;

    // This hash works on FAT sequence index.  The Directory
    // Hash is identical in structure.  This method allows very
    // rapid table searches of the directory file and the FAT

    NWLockFat(volume);
    hash = (lru->BlockIndex & (volume->FATBlockHashLimit - 1));
    HashTable = (LRU_HASH_LIST *) volume->FATBlockHash;
    if (HashTable)
    {
       if (!HashTable[hash].head)
       {
	  HashTable[hash].head = lru;
	  HashTable[hash].tail = lru;
	  lru->hashNext = lru->hashPrior = 0;
       }
       else
       {
	  HashTable[hash].tail->hashNext = lru;
	  lru->hashNext = 0;
	  lru->hashPrior = HashTable[hash].tail;
	  HashTable[hash].tail = lru;
       }
       NWUnlockFat(volume);
       return 0;
    }
    NWUnlockFat(volume);
    return -1;
}

ULONG RemoveFATHash(VOLUME *volume, MIRROR_LRU *lru)
{
    register ULONG hash;
    register LRU_HASH_LIST *HashTable;

    NWLockFat(volume);
    hash = (lru->BlockIndex & (volume->FATBlockHashLimit - 1));
    HashTable = (LRU_HASH_LIST *) volume->FATBlockHash;
    if (HashTable)
    {
       if (HashTable[hash].head == lru)
       {
	  HashTable[hash].head = lru->hashNext;
	  if (HashTable[hash].head)
	     HashTable[hash].head->hashPrior = NULL;
	  else
	     HashTable[hash].tail = NULL;
       }
       else
       {
	  lru->hashPrior->hashNext = lru->hashNext;
	  if (lru != HashTable[hash].tail)
	     lru->hashNext->hashPrior = lru->hashPrior;
	  else
	     HashTable[hash].tail = lru->hashPrior;
       }
       NWUnlockFat(volume);
       return 0;
    }
    NWUnlockFat(volume);
    return -1;
}

