
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  SUPER.C
*   DESCRIP  :  FENRIS VFS Super Block Module for Linux*   DATE     :  November 16, 1998
*
*
***************************************************************************/

#define NWFSMOD_VERSION_INFO 1
#include "globals.h"


#if (LINUX_24)

int nwfs_remount(struct super_block *sb, int *flags, char *data)
{
    return 0;
}

struct super_block *nwfs_read_super(struct super_block *sb, void *data, int silent)
{
    register VOLUME *volume;
    register BYTE *name;
    kdev_t dev = sb->s_dev;

#if (VERBOSE)
    NWFSPrint("read super\n");
#endif
    
    if (!data)
    {
       NWFSPrint("nwfs:  missing VOLUME_NAME\n");
       NWFSPrint("USAGE: mount [/dev/dev_name] [/mount_point] -t nwfs -o VOLUME_NAME\n");
       NWFSPrint("       where VOLUME_NAME is the name of a NetWare file system Volume\n");
       NWFSPrint("       invoke NWVOL for a listing of detected NetWare Volumes in\n");
       NWFSPrint("       this system.\n\n");
       NWFSPrint("       FENRIS Utilities for Linux are:\n");
       NWFSPrint("       NWVOL            - List detected NetWare volumes\n");
       NWFSPrint("       NWDISK/NWCONFIG  - File System/Partition Config Utility\n");
       sb->s_dev = 0;
       return NULL;
    }

    NWFSVolumeScan();
    name = (BYTE *) data;
    volume = MountHandledVolume(name);
    if (!volume)
    {
       NWFSPrint("nwfs:  error mounting volume %s\n", name);
       sb->s_dev = 0;
       return NULL;
    }

    sb->s_blocksize = 512;
    sb->s_blocksize_bits = LINUX_512_SHIFT;
    sb->s_magic = FENRIS_LINUX_FS_ID;
    sb->s_dev = dev;
    sb->s_op = &nwfs_sops;
    sb->u.generic_sbp = (void *) volume;

    sb->s_root = d_alloc_root(iget(sb, 0));
    if (!sb->s_root)
    {
       NWFSPrint("nwfs:  get root inode failed\n");
       sb->s_dev = 0;
       return NULL;
    }
    return sb;
}

void nwfs_put_super(struct super_block *sb)
{
    register VOLUME *volume;
    register ULONG retCode;

#if (VERBOSE)
    NWFSPrint("put super\n");
#endif
    
    sb->s_dev = 0;
    volume = (VOLUME *) sb->u.generic_sbp;
    sb->u.generic_sbp = 0;

    retCode = DismountVolumeByHandle(volume);
    if (retCode)
       NWFSPrint("nwfs:  errors dismounting volume %s", volume->VolumeName);

    return;
}

int nwfs_statfs(struct super_block *sb, struct statfs *buf)
{
    register VOLUME *volume = (VOLUME *) sb->u.generic_sbp;

    buf->f_type = FENRIS_LINUX_FS_ID;
    buf->f_bsize = 512;
    buf->f_blocks = (volume->VolumeClusters * volume->SectorsPerCluster);
    buf->f_bfree = (volume->VolumeFreeClusters * volume->SectorsPerCluster);
    buf->f_bavail = (volume->VolumeFreeClusters * volume->SectorsPerCluster);
    buf->f_namelen = 255;

    return 0;
}

#endif


#if (LINUX_22)

int nwfs_remount(struct super_block *sb, int *flags, char *data)
{
    return 0;
}

struct super_block *nwfs_read_super(struct super_block *sb, void *data, int silent)
{
    register VOLUME *volume;
    register BYTE *name;
    kdev_t dev = sb->s_dev;

    if (!data)
    {
       NWFSPrint("nwfs:  missing VOLUME_NAME\n");
       NWFSPrint("USAGE: mount [/dev/dev_name] [/mount_point] -t nwfs -o VOLUME_NAME\n");
       NWFSPrint("       where VOLUME_NAME is the name of a NetWare file system Volume\n");
       NWFSPrint("       invoke NWVOL for a listing of detected NetWare Volumes in\n");
       NWFSPrint("       this system.\n\n");
       NWFSPrint("       FENRIS Utilities for Linux are:\n");
       NWFSPrint("       NWVOL            - List detected NetWare volumes\n");
       NWFSPrint("       NWDISK/NWCONFIG  - File System/Partition Config Utility\n");
       sb->s_dev = 0;
       return NULL;
    }

    MOD_INC_USE_COUNT;
    lock_super(sb);

    NWFSVolumeScan();
    name = (BYTE *) data;
    volume = MountHandledVolume(name);
    if (!volume)
    {
       NWFSPrint("nwfs:  error mounting volume %s\n", name);
       sb->s_dev = 0;
       unlock_super(sb);
       MOD_DEC_USE_COUNT;
       return NULL;
    }

    sb->s_blocksize = 512;
    sb->s_blocksize_bits = LINUX_512_SHIFT;
    sb->s_magic = FENRIS_LINUX_FS_ID;
    sb->s_dev = dev;
    sb->s_op = &nwfs_sops;
    sb->u.generic_sbp = (void *) volume;

    sb->s_root = d_alloc_root(iget(sb, 0), NULL);
    if (!sb->s_root)
    {
       NWFSPrint("nwfs:  get root inode failed\n");
       sb->s_dev = 0;
       unlock_super(sb);
       MOD_DEC_USE_COUNT;
       return NULL;
    }
    unlock_super(sb);
    return sb;
}

void nwfs_put_super(struct super_block *sb)
{
    register VOLUME *volume;
    register ULONG retCode;

    sb->s_dev = 0;
    volume = (VOLUME *) sb->u.generic_sbp;
    sb->u.generic_sbp = 0;

    retCode = DismountVolumeByHandle(volume);
    if (retCode)
       NWFSPrint("nwfs:  errors dismounting volume %s", volume->VolumeName);

    MOD_DEC_USE_COUNT;
    return;
}

int nwfs_statfs(struct super_block *sb, struct statfs *buf, int bufsiz)
{
    register VOLUME *volume = (VOLUME *) sb->u.generic_sbp;
    struct statfs tmp;

    tmp.f_type = FENRIS_LINUX_FS_ID;
    tmp.f_bsize = 512;
    tmp.f_blocks = (volume->VolumeClusters * volume->SectorsPerCluster);
    tmp.f_bfree = (volume->VolumeFreeClusters * volume->SectorsPerCluster);
    tmp.f_bavail = (volume->VolumeFreeClusters * volume->SectorsPerCluster);
    tmp.f_files = 0;
    tmp.f_ffree = 0;
    tmp.f_namelen = 255;
    return copy_to_user(buf, &tmp, bufsiz) ? -EFAULT : 0;
}

#endif



#if (LINUX_20)

int nwfs_remount(struct super_block *sb, int *flags, char *data)
{
    return 0;
}

struct super_block *nwfs_read_super(struct super_block *sb, void *data, int silent)
{
    register VOLUME *volume;
    register BYTE *name;
    kdev_t dev = sb->s_dev;

    if (!data)
    {
       NWFSPrint("nwfs:  missing VOLUME_NAME\n");
       NWFSPrint("USAGE: mount [/dev/dev_name] [/mount_point] -t nwfs -o VOLUME_NAME\n");
       NWFSPrint("       where VOLUME_NAME is the name of a NetWare file system Volume\n");
       NWFSPrint("       invoke NWVOL for a listing of detected NetWare Volumes in\n");
       NWFSPrint("       this system.\n\n");
       NWFSPrint("       FENRIS Utilities for Linux are:\n");
       NWFSPrint("       NWVOL            - List detected NetWare volumes\n");
       NWFSPrint("       NWDISK/NWCONFIG  - File System/Partition Config Utility\n");
       sb->s_dev = 0;
       return NULL;
    }

    NWFSVolumeScan();
    name = (BYTE *) data;
    volume = MountHandledVolume(name);
    if (!volume)
    {
       NWFSPrint("nwfs:  error mounting volume %s\n", name);
       sb->s_dev = 0;
       return NULL;
    }

    MOD_INC_USE_COUNT;

    lock_super(sb);

    sb->s_blocksize = 512;
    sb->s_blocksize_bits = LINUX_512_SHIFT;
    sb->s_magic = FENRIS_LINUX_FS_ID;
    sb->s_dev = dev;
    sb->s_op = &nwfs_sops;
    sb->u.generic_sbp = (void *) volume;

    unlock_super(sb);

    sb->s_mounted = iget(sb, 0);
    if (!sb->s_mounted)
    {
       NWFSPrint("nwfs:  get root inode failed\n");
       sb->s_dev = 0;
       MOD_DEC_USE_COUNT;
       return NULL;
    }

    sb->s_mounted->i_atime = sb->s_mounted->i_mtime = NWFSGetSystemTime();
    sb->s_mounted->i_ctime = NWFSGetSystemTime();

    return sb;

}

void nwfs_put_super(struct super_block *sb)
{
    register VOLUME *volume;
    register ULONG retCode;

    lock_super(sb);

    sb->s_dev = 0;
    volume = (VOLUME *) sb->u.generic_sbp;
    sb->u.generic_sbp = 0;

    unlock_super(sb);

    retCode = DismountVolumeByHandle(volume);
    if (retCode)
       NWFSPrint("nwfs:  errors dismounting volume %s", volume->VolumeName);

    MOD_DEC_USE_COUNT;

    return;
}

void nwfs_statfs(struct super_block *sb, struct statfs *buf, int bufsiz)
{
    register VOLUME *volume = (VOLUME *) sb->u.generic_sbp;
    struct statfs tmp;

    tmp.f_type = FENRIS_LINUX_FS_ID;
    tmp.f_bsize = 512;
    tmp.f_blocks = (volume->VolumeClusters * volume->SectorsPerCluster);
    tmp.f_bfree = (volume->VolumeFreeClusters * volume->SectorsPerCluster);
    tmp.f_bavail = (volume->VolumeFreeClusters * volume->SectorsPerCluster);
    tmp.f_files = 0;
    tmp.f_ffree = 0;
    tmp.f_namelen = 255;
    NWFSCopyToUserSpace(buf, &tmp, bufsiz);

}

int nwfs_register_symbols(void)
{
    return 0;
}

#endif

